@@ -1,12 +1,11 @@
 //
 // Copyright(C) 2001-2005 Simon Howard
 //
-// This program is free software; you can redistribute it and/or modify it
-// under the terms of the GNU General Public License as published by the
-// Free Software Foundation; either version 2 of the License, or (at your
-// option) any later version. This program is distributed in the hope that
-// it will be useful, but WITHOUT ANY WARRANTY; without even the implied
-// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
+// You can redistribute and/or modify this program under the terms of the
+// GNU General Public License version 2 as published by the Free Software
+// Foundation, or any later version. This program is distributed WITHOUT
+// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
+// FITNESS FOR A PARTICULAR PURPOSE.
 //
 //
 // SDL Video Code
@@ -24,25 +23,53 @@
 
 #include "video.h"
 #include "sw.h"
+#include "swinit.h"
+#include "swmain.h"
 
-// lcd mode to emulate my old laptop i used to play sopwith on :)
+#define INPUT_BUFFER_LEN 32
 
-//#define LCD
-
-static SDL_Color cga_pal[] = {
-#ifdef LCD
-	{213, 226, 138}, {150, 160, 150},
-	{120, 120, 160}, {0, 20, 200},
-#else
-	{0, 0, 0}, {0, 255, 255},
-	{255, 0, 255}, {255, 255, 255},
-#endif
+struct palette {
+	char name[13]; // Up to 12 characters will display correctly on the menu
+	SDL_Color color[4];
 };
 
-bool vid_fullscreen = false;
+static const struct palette video_palettes[] = {
+	{"CGA 1", 		// CGA black, cyan, magenta, white (Sopwith's default color scheme)
+		{{0, 0, 0}, {0, 255, 255}, {255, 0, 255}, {255, 255, 255}}},
+	{"CGA 2", 		// CGA black, red, green, yellow
+		{{0, 0, 0}, {0, 255, 0}, {255, 0, 0}, {255, 255, 0}}},
+	{"CGA 3", 		// CGA black, cyan, red, white (aka CGA mode 5)
+		{{0, 0, 0}, {0, 255, 255}, {255, 0, 0}, {255, 255, 255}}},
+	{"Mono Amber",   // Shades of amber from a monochrome CGA display
+		{{0, 0, 0}, {255, 170, 16}, {242, 125, 0}, {255, 226, 52}}},
+	{"Mono Green", 	// Shades of green from a monochrome CGA display
+		{{0, 0, 0}, {12, 238, 56}, {8, 202, 48}, {49, 253, 90}}},
+	{"Mono Grey", 	// Shades of grey from a monochrome CGA display
+		{{0, 0, 0}, {222, 222, 210}, {182, 186, 182}, {255, 255, 255}}},
+	{"Tosh LCD 1",		// Toshiba laptop with STN panel
+		{{213, 226, 138}, {150, 160, 150}, {120, 120, 160}, {0, 20, 200}}},
+	{"Tosh LCD 2",		// Toshiba laptop with STN panel, reversed
+		{{0, 20, 200}, {120, 120, 160}, {150, 160, 150}, {213, 226, 138}}},
+	{"Tosh LCD 3",		// Toshiba T1000 with no backlight
+		{{0x72, 0x88, 0x79}, {0x4b, 0x6e, 0x75},
+		 {0x42, 0x5a, 0x75}, {0x27, 0x46, 0x6d}}},
+	{"IBM LCD",  // IBM PC Convertible
+		{{0x6b, 0x85, 0x88}, {0x56, 0x6b, 0x6e},
+		 {0x42, 0x52, 0x54}, {0x2e, 0x39, 0x3b}}},
+	{"Tandy LCD", // Tandy 1100FD
+		{{0x48, 0xad, 0x68}, {0x36, 0x8c, 0x61},
+		 {0x24, 0x6c, 0x5a}, {0x13, 0x4a, 0x54}}},
+	{"Gas Plasma",
+		{{0x7d, 0x1b, 0x02}, {0xd3, 0x41, 0x00},
+		 {0xa8, 0x2e, 0x01}, {0xfe, 0x54, 0x00}}},
+	// Palette from swgrapha.c in the original source, the colors
+	// that would have appeared in the Atari port.
+	{"Atari",
+		{{0x00, 0x00, 0x00}, {0x00, 0x77, 0xff},
+		 {0xff, 0x00, 0x00}, {0xff, 0xff, 0xff}}},
+};
 
-extern unsigned char *vid_vram;
-extern unsigned int vid_pitch;
+extern bool isNetworkGame(void);
 
 int keybindings[NUM_KEYS] = {
 	0,                    // KEY_UNKNOWN
@@ -59,11 +86,14 @@ int keybindings[NUM_KEYS] = {
 	SDL_SCANCODE_S,       // KEY_SOUND
 };
 
+bool vid_fullscreen = false;
 static int ctrlbreak = 0;
-static bool initted = 0;
+static bool initted = false;
 static SDL_Window *window = NULL;
 static uint32_t pixel_format;
 static SDL_Renderer *renderer;
+static SDL_Keysym input_buffer[INPUT_BUFFER_LEN];
+static int input_buffer_head = 0, input_buffer_tail = 0;
 
 // Maximum number of pixels to use for intermediate scale buffer.
 static int max_scaling_buffer_pixels = 16000000;
@@ -79,16 +109,16 @@ static SDL_Surface *argbbuffer = NULL;
 static SDL_Texture *texture = NULL;
 static SDL_Texture *texture_upscaled = NULL;
 
-
 // convert a sopsym_t into a surface
 #define ICON_SCALE 4
-static SDL_Surface *surface_from_sopsym(sopsym_t *sym)
+static SDL_Surface *SurfaceFromSopsym(sopsym_t *sym)
 {
+	const SDL_Color *pal = video_palettes[0].color;
 	SDL_Surface *surface = SDL_CreateRGBSurface(
 		0, sym->w * ICON_SCALE, sym->h * ICON_SCALE, 32,
 		0xff000000, 0x00ff0000, 0x0000ff00, 0x000000ff);
 
-	unsigned char *src;
+	uint8_t *src;
 	uint32_t *dst;
 	int x, y, sx, sy;
 
@@ -106,13 +136,13 @@ static SDL_Surface *surface_from_sopsym(sopsym_t *sym)
 		dst = (uint32_t *) ((uint8_t *) surface->pixels +
 		                    y * surface->pitch);
 		for (x = 0; x < surface->w; ++x) {
-			SDL_Color *p;
+			const SDL_Color *p;
 			sx = x / ICON_SCALE;
 			if (src[sx] == 0) {
 				dst[x] = 0;
 				continue;
 			}
-			p = &cga_pal[src[sx]];
+			p = &pal[src[sx]];
 			dst[x] = (p->r << 24) | (p->g << 16)
 			       | (p->b << 8) | 0xff;
 		}
@@ -123,8 +153,6 @@ static SDL_Surface *surface_from_sopsym(sopsym_t *sym)
 	return surface;
 }
 
-// 2x scale
-
 void Vid_Update(void)
 {
 	static SDL_Rect blit_rect = { 0, 0, SCR_WDTH, SCR_HGHT };
@@ -160,7 +188,7 @@ void Vid_Update(void)
 	SDL_LockSurface(screenbuf);
 }
 
-static bool is_special_day(void)
+static bool IsSpecialDay(void)
 {
 	time_t now = time(NULL);
 	struct tm *t = localtime(&now);
@@ -179,18 +207,18 @@ static bool is_special_day(void)
 	 || (t->tm_mon == 3 && t->tm_mday == 24);
 }
 
-static void set_icon(void)
+static void SetIcon(void)
 {
 	SDL_Surface *icon;
 	sopsym_t *sym;
 
-	if (is_special_day()) {
-		sym = symbol_plane[1][0];
+	if (IsSpecialDay()) {
+		sym = &symbol_plane[0].sym[4];
 	} else {
-		sym = symbol_plane[0][0];
+		sym = &symbol_plane[0].sym[0];
 	}
 
-	icon = surface_from_sopsym(sym);
+	icon = SurfaceFromSopsym(sym);
 	if (icon == NULL) {
 		return;
 	}
@@ -210,8 +238,8 @@ static void LimitTextureSize(int *w_upscale, int *h_upscale)
 
 	// Query renderer and limit to maximum texture dimensions of hardware:
 	if (SDL_GetRendererInfo(renderer, &rinfo) != 0) {
-		error_exit("CreateUpscaledTexture: SDL_GetRendererInfo() "
-		           "call failed: %s", SDL_GetError());
+		ErrorExit("CreateUpscaledTexture: SDL_GetRendererInfo() "
+		          "call failed: %s", SDL_GetError());
 	}
 
 	while (*w_upscale * SCR_WDTH > rinfo.max_texture_width) {
@@ -223,10 +251,10 @@ static void LimitTextureSize(int *w_upscale, int *h_upscale)
 
 	if ((*w_upscale < 1 && rinfo.max_texture_width > 0)
 	 || (*h_upscale < 1 && rinfo.max_texture_height > 0)) {
-		error_exit("CreateUpscaledTexture: Can't create a "
-		           "texture big enough for the whole screen! "
-		           "Maximum texture size %dx%d",
-		           rinfo.max_texture_width, rinfo.max_texture_height);
+		ErrorExit("CreateUpscaledTexture: Can't create a "
+		          "texture big enough for the whole screen! "
+		          "Maximum texture size %dx%d",
+		          rinfo.max_texture_width, rinfo.max_texture_height);
 	}
 
 	// We limit the amount of texture memory used for the intermediate buffer,
@@ -235,10 +263,10 @@ static void LimitTextureSize(int *w_upscale, int *h_upscale)
 	// huge textures, so the user can use this to reduce the maximum texture
 	// size if desired.
 	if (max_scaling_buffer_pixels < SCR_WDTH * SCR_HGHT) {
-		error_exit("CreateUpscaledTexture: max_scaling_buffer_"
-		           "pixels too small to create a texture buffer:"
-		           " %d < %d", max_scaling_buffer_pixels,
-		           SCR_WDTH * SCR_HGHT);
+		ErrorExit("CreateUpscaledTexture: max_scaling_buffer_"
+		          "pixels too small to create a texture buffer:"
+		          " %d < %d", max_scaling_buffer_pixels,
+		          SCR_WDTH * SCR_HGHT);
 	}
 
 	while (*w_upscale * *h_upscale * SCR_WDTH * SCR_HGHT
@@ -270,7 +298,7 @@ static void CreateUpscaledTexture(int force)
 	// real world pixels, which are not necessarily equivalent to the
 	// screen's window size (because of highdpi).
 	if (SDL_GetRendererOutputSize(renderer, &w, &h) != 0) {
-		error_exit("failed to get renderer size: %s", SDL_GetError());
+		ErrorExit("failed to get renderer size: %s", SDL_GetError());
 	}
 
 	// When the screen or window dimensions do not match the aspect ratio
@@ -287,16 +315,9 @@ static void CreateUpscaledTexture(int force)
 	// Pick texture size the next integer multiple of the screen dimensions.
 	// If one screen dimension matches an integer multiple of the original
 	// resolution, there is no need to overscale in this direction.
-	w_upscale = (w + SCR_WDTH - 1) / SCR_WDTH;
-	h_upscale = (h + SCR_HGHT - 1) / SCR_HGHT;
-
 	// Minimum texture dimensions of 320x200.
-	if (w_upscale < 1) {
-		w_upscale = 1;
-	}
-	if (h_upscale < 1) {
-		h_upscale = 1;
-	}
+	w_upscale = clamp_min(1, (w + SCR_WDTH - 1) / SCR_WDTH);
+	h_upscale = clamp_min(1, (h + SCR_HGHT - 1) / SCR_HGHT);
 
 	LimitTextureSize(&w_upscale, &h_upscale);
 
@@ -360,8 +381,8 @@ static void Vid_SetMode(void)
 	int bpp;
 
 	if (SDL_Init(SDL_INIT_VIDEO) < 0) {
-		error_exit("Unable to initialize video subsystem: %s",
-		           SDL_GetError());
+		ErrorExit("Unable to initialize video subsystem: %s",
+		          SDL_GetError());
 	}
 	srand(time(NULL));
 
@@ -373,20 +394,22 @@ static void Vid_SetMode(void)
 		flags |= SDL_WINDOW_FULLSCREEN_DESKTOP;
 	}
 
-	window = SDL_CreateWindow(PACKAGE_STRING, SDL_WINDOWPOS_UNDEFINED,
-	                          SDL_WINDOWPOS_UNDEFINED, w, h, flags);
+	window = SDL_CreateWindow(PACKAGE_STRING, SDL_WINDOWPOS_CENTERED,
+	                          SDL_WINDOWPOS_CENTERED, w, h, flags);
 
 	if (window == NULL) {
-		error_exit("Failed to open SDL window: %s", SDL_GetError());
+		ErrorExit("Failed to open SDL window: %s", SDL_GetError());
 	}
 
-        pixel_format = SDL_GetWindowPixelFormat(window);
+	SDL_StopTextInput();
+
+	pixel_format = SDL_GetWindowPixelFormat(window);
 
 	for (n = 0; n < NUM_KEYS; ++n) {
 		keysdown[n] = 0;
 	}
 
-	set_icon();
+	SetIcon();
 	SDL_ShowCursor(0);
 
 	renderer_flags = SDL_RENDERER_PRESENTVSYNC;
@@ -450,7 +473,7 @@ void Vid_Shutdown(void)
 
 	SDL_FreeSurface(screenbuf);
 
-	initted = 0;
+	initted = false;
 }
 
 void Vid_Init(void)
@@ -467,10 +490,11 @@ void Vid_Init(void)
 	                                 0, 0, 0, 0);
 	vid_vram = screenbuf->pixels;
 	vid_pitch = screenbuf->pitch;
-	SDL_SetPaletteColors(screenbuf->format->palette, cga_pal, 0,
-	                     sizeof(cga_pal) / sizeof(*cga_pal));
+	SDL_SetPaletteColors(screenbuf->format->palette,
+	                     video_palettes[0].color, 0,
+	                     arrlen(video_palettes[0].color));
 
-	initted = 1;
+	initted = true;
 
 	atexit(Vid_Shutdown);
 
@@ -487,15 +511,28 @@ void Vid_Reset(void)
 	Vid_SetMode();
 
 	// need to redraw buffer to screen
-
 	Vid_Update();
 }
 
-#define INPUT_BUFFER_LEN 32
-static SDL_Keysym input_buffer[INPUT_BUFFER_LEN];
-static int input_buffer_head = 0, input_buffer_tail = 0;
+void Vid_SetVideoPalette(int palette)
+{
+	SDL_SetPaletteColors(screenbuf->format->palette,
+	                     video_palettes[palette].color, 0,
+	                     arrlen(video_palettes[palette].color));
+	Vid_Update();
+}
 
-static void input_buffer_push(SDL_Keysym c)
+const char* Vid_GetVideoPaletteName(int palette)
+{
+	return video_palettes[palette].name;
+}
+
+int Vid_GetNumVideoPalettes(void)
+{
+	return arrlen(video_palettes);
+}
+
+static void InputBufferPush(SDL_Keysym c)
 {
 	int tail_next = (input_buffer_tail + 1) % INPUT_BUFFER_LEN;
 	if (tail_next == input_buffer_head) {
@@ -505,7 +542,7 @@ static void input_buffer_push(SDL_Keysym c)
 	input_buffer_tail = tail_next;
 }
 
-static SDL_Keysym input_buffer_pop(void)
+static SDL_Keysym InputBufferPop(void)
 {
 	SDL_Keysym result;
 
@@ -520,7 +557,7 @@ static SDL_Keysym input_buffer_pop(void)
 	return result;
 }
 
-static sopkey_t translate_scancode(int sdl_scancode)
+static enum gamekey TranslateScancode(int sdl_scancode)
 {
 	int i;
 
@@ -546,83 +583,127 @@ static bool IsSpecialKey(SDL_Keysym *k) {
 	}
 }
 
-static void getevents(void)
+static bool CtrlDown(void)
+{
+	return (SDL_GetModState() & KMOD_CTRL) != 0;
+}
+
+static bool AltDown(void)
+{
+	return (SDL_GetModState() & KMOD_ALT) != 0;
+}
+
+static void CtrlKeyPress(SDL_Keycode k)
+{
+	switch (k) {
+#ifndef NO_EXIT
+	case SDLK_c:
+	case SDLK_PAUSE:
+		++ctrlbreak;
+		if (ctrlbreak >= 3) {
+			fprintf(stderr,
+				"user aborted with 3 ^C's\n");
+			exit(-1);
+		}
+		break;
+#endif
+	case SDLK_r:
+		if (!isNetworkGame()) {
+			gamenum = starting_level;
+			swinitlevel();
+		}
+		break;
+	case SDLK_q:
+		if (!isNetworkGame()) {
+			swrestart();
+		}
+		break;
+	default:
+		break;
+	}
+}
+
+static void KeyDown(SDL_KeyboardEvent *event)
+{
+	enum gamekey translated;
+
+	if (CtrlDown()) {
+		CtrlKeyPress(event->keysym.sym);
+		return;
+	} else if (AltDown() && (event->keysym.sym == SDLK_RETURN
+	                      || event->keysym.sym == SDLK_KP_ENTER)) {
+#ifndef NO_FULLSCREEN
+		vid_fullscreen = !vid_fullscreen;
+		Vid_Reset();
+#endif
+		return;
+	} else if (event->keysym.sym == SDLK_KP_ENTER
+	        && SDL_IsTextInputActive()) {
+		SDL_Keysym fake = {0};
+		fake.sym = '\n';
+		fake.scancode = SDL_SCANCODE_UNKNOWN;
+		InputBufferPush(fake);
+		return;
+	} else if (!SDL_IsTextInputActive() || IsSpecialKey(&event->keysym)) {
+		InputBufferPush(event->keysym);
+	}
+
+	translated = TranslateScancode(event->keysym.scancode);
+	if (translated != KEY_UNKNOWN) {
+		keysdown[translated] |= 3;
+	}
+}
+
+static void KeyUp(SDL_KeyboardEvent *event)
+{
+	enum gamekey translated = TranslateScancode(event->keysym.scancode);
+	if (translated != KEY_UNKNOWN) {
+		keysdown[translated] &= ~1;
+	}
+}
+
+static void GetEvents(void)
 {
 	SDL_Event event;
-	static bool ctrldown = 0, altdown = 0;
-	int need_redraw = 0;
+	bool need_redraw = false;
 	int i;
-	sopkey_t translated;
 
 	while (SDL_PollEvent(&event)) {
 		switch (event.type) {
 		case SDL_KEYDOWN:
-			if (event.key.keysym.sym == SDLK_LALT) {
-				altdown = 1;
-			} else if (event.key.keysym.sym == SDLK_LCTRL
-			        || event.key.keysym.sym == SDLK_RCTRL) {
-				ctrldown = 1;
-			} else if (ctrldown &&
-			           (event.key.keysym.sym == SDLK_c ||
-			            event.key.keysym.sym == SDLK_PAUSE)) {
-				++ctrlbreak;
-				if (ctrlbreak >= 3) {
-					fprintf(stderr,
-						"user aborted with 3 ^C's\n");
-					exit(-1);
-				}
-			} else if (event.key.keysym.sym == SDLK_RETURN) {
-				if (altdown) {
-					vid_fullscreen = !vid_fullscreen;
-					Vid_Reset();
-					continue;
-				}
-			}
-			if (!SDL_IsTextInputActive()
-			 || IsSpecialKey(&event.key.keysym)) {
-				input_buffer_push(event.key.keysym);
-			}
-			translated = translate_scancode(
-				event.key.keysym.scancode);
-			if (translated != KEY_UNKNOWN) {
-				keysdown[translated] |= 3;
-			}
+			KeyDown(&event.key);
 			break;
 
 		case SDL_KEYUP:
-			if (event.key.keysym.sym == SDLK_LALT) {
-				altdown = 0;
-			} else if (event.key.keysym.sym == SDLK_LCTRL
-			        || event.key.keysym.sym == SDLK_RCTRL) {
-				ctrldown = 0;
-			} else {
-				translated = translate_scancode(
-					event.key.keysym.scancode);
-				if (translated != KEY_UNKNOWN) {
-					keysdown[translated] &= ~1;
-				}
-			}
+			KeyUp(&event.key);
 			break;
 
 		case SDL_TEXTINPUT:
 			for (i = 0; event.text.text[i] != '\0'; ++i) {
 				char c = event.text.text[i];
-				SDL_Keysym fake;
+				SDL_Keysym fake = {0};
 				if (c >= 0x80) {
 					continue;
 				}
 				fake.sym = c;
 				fake.scancode = SDL_SCANCODE_UNKNOWN;
-				input_buffer_push(fake);
+				InputBufferPush(fake);
 			}
 			break;
 
 		case SDL_WINDOWEVENT:
-			// When we get one of these events, we redraw
-			// the screen immediately, as we may be in a
-			// menu waiting for a keypress.
-			need_redraw = 1;
-			break;
+			switch (event.window.event) {
+			case SDL_WINDOWEVENT_CLOSE:
+				exit(0);
+				break;
+
+			default:
+				// When we get one of these events, we redraw
+				// the screen immediately, as we may be in a
+				// menu waiting for a keypress.
+				need_redraw = true;
+				break;
+			}
 		}
 	}
 
@@ -644,16 +725,16 @@ void Vid_StopTextInput(void)
 int Vid_GetKey(void)
 {
 	SDL_Keysym k;
-	getevents();
-	k = input_buffer_pop();
+	GetEvents();
+	k = InputBufferPop();
 	return k.scancode;
 }
 
 int Vid_GetChar(void)
 {
 	int result;
-	getevents();
-	result = input_buffer_pop().sym;
+	GetEvents();
+	result = InputBufferPop().sym;
 	if (result == '\r') {
 		result = '\n';
 	}
@@ -662,7 +743,7 @@ int Vid_GetChar(void)
 
 bool Vid_GetCtrlBreak(void)
 {
-	getevents();
+	GetEvents();
 	return ctrlbreak;
 }
 
@@ -695,7 +776,7 @@ int isatty(int fd)
 }
 #endif
 
-void error_exit(char *s, ...)
+void ErrorExit(char *s, ...)
 {
 	static char buf[128];
 	va_list args;
