@@ -2,12 +2,11 @@
 // Copyright(C) 1984-2000 David L. Clark
 // Copyright(C) 2001-2005 Simon Howard
 //
-// This program is free software; you can redistribute it and/or modify it
-// under the terms of the GNU General Public License as published by the
-// Free Software Foundation; either version 2 of the License, or (at your
-// option) any later version. This program is distributed in the hope that
-// it will be useful, but WITHOUT ANY WARRANTY; without even the implied
-// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
+// You can redistribute and/or modify this program under the terms of the
+// GNU General Public License version 2 as published by the Free Software
+// Foundation, or any later version. This program is distributed WITHOUT
+// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
+// FITNESS FOR A PARTICULAR PURPOSE.
 //
 //
 //        swinit   -      SW initialization
@@ -16,6 +15,7 @@
 #include <stdio.h>
 #include <string.h>
 #include <time.h>
+#include <assert.h>
 
 #include "pcsound.h"
 #include "timer.h"
@@ -23,8 +23,9 @@
 
 #include "sw.h"
 #include "swasynio.h"
+#include "swcollsn.h"
 #include "swconf.h"
-#include "swdisp.h"
+#include "swgrpha.h"
 #include "swinit.h"
 #include "swgames.h"
 #include "swmain.h"
@@ -35,62 +36,52 @@
 #include "swsymbol.h"
 #include "swtitle.h"
 
+static const original_ob_t *orig_planes[MAX_PLYR * 2];
 GRNDTYPE *ground;
 
-static int have_savescore = 0;
+static bool have_savescore = false;
 static score_t savescore;		/* save players score on restart  */
-static int starting_level = 0;
+int starting_level = 0;
 
-static char helptxt[] =
+static const char *helptxt =
 "\n"
 PACKAGE_STRING "\n"
 "Copyright (C) 1984, 1985, 1987 BMB Compuscience\n"
 "Copyright (C) 1984-2000 David L. Clark\n"
-"Copyright (C) 2001-2022 Simon Howard, Jesse Smith\n"
+"Copyright (C) 2001-2024 Simon Howard, Jesse Smith\n"
 "Licensed under the GNU GPL v2.\n"
 "\n"
 "Usage:  sopwith [options]\n"
 "The options are:\n"
-"        -n :  novice single player\n"
-"        -s :  single player\n"
-"        -c :  single player against computer\n"
-"        -q :  begin game with sound off\n"
-"        -g#:  start at level #\n"
-"        -e :  turn off big explosions\n"
+"        -n : novice single player\n"
+"        -s : single player\n"
+"        -c : single player against computer\n"
+"       -g# : start at level #\n"
 "\n"
-"Video:\n"
-"        -f    :  fullscreen\n"
+"        -e : turn off big explosions\n"
+"        -f : fullscreen\n"
+" -m <file> : load new mission from file\n"
+"        -q : begin game with sound off\n"
 "\n"
 #ifdef TCPIP
 "Networking: \n"
-"        -l    :  listen for connection\n"
-" -j <host>    :  connect to a listening host\n"
+" -j <host> : connect to a listening host\n"
+"        -l : listen for connection\n"
+" -p <port> : use alternative TCP port\n"
 #endif
 ;
 
 static void initobjs(void)
 {
-	OBJECTS *ob;
-	int i;
-
 	topobj.ob_xnext = topobj.ob_next = &botobj;
 	botobj.ob_xprev = botobj.ob_prev = &topobj;
 	topobj.ob_x = -32767;
 	botobj.ob_x = 32767;
 
 	objbot = objtop = deltop = delbot = NULL;
-	objfree = ob = nobjects;
-
-	for (i = 0; i < MAX_OBJS; ++i) {
-		ob->ob_next = ob + 1;
-		ob->ob_index = i;
-		++ob;
-	}
-
-	(ob-1)->ob_next = NULL;
+	objfree = NULL;
 }
 
-
 void initgrnd(void)
 {
 	free(ground);
@@ -113,21 +104,23 @@ void initdisp(bool reset)
 	}
 }
 
-
 //
 // object creation
 //
 
-
-static void initplanescore (score_t *score)
+static void initscore(score_t *score)
 {
-	score->planekills = 0;
-	score->medals = 0x0000;
-	score->valour = 0;
-	score->killscore = 0;
-	score->landings = 0;
-	score->medals_nr = 0;
-	score->ribbons_nr = 0;
+	memset(score, 0, sizeof(score_t));
+
+	if (have_savescore) {
+		*score = savescore;
+		have_savescore = false;
+	}
+}
+
+static void initflightscore(flight_score_t *score)
+{
+	memset(score, 0, sizeof(flight_score_t));
 }
 
 #define SERVICE_KILLSCORE 5
@@ -135,95 +128,98 @@ static void initplanescore (score_t *score)
 #define VALOUR_PRELIMIT 175
 #define VALOUR_LIMIT 250
 
-static void give_medal(OBJECTS *ob, int medal_id)
+static bool HaveMedal(int *medals, int medals_nr, int medal_id)
 {
-	ob->ob_score.medalslist[ob->ob_score.medals_nr++] = medal_id;
+	int i;
+
+	for (i = 0; i < medals_nr; i++) {
+		if (medals[i] == medal_id) {
+			return true;
+		}
+	}
+
+	return false;
 }
 
-static void give_ribbon(OBJECTS *ob, int ribbon_id)
-{
-	ob->ob_score.ribbons[ob->ob_score.ribbons_nr++] = ribbon_id;
-}
-
-static void get_awards(OBJECTS *ob)
-{
-	score_t *lsc = &ob->ob_lastscore;
-	score_t *sc = &ob->ob_score;
-
-	if (!(lsc->medals & MEDAL_PURPLEHEART)) {
-		if (ob->ob_state == WOUNDED
-		    || ob->ob_state == WOUNDSTALL) {
-			sc->medals |= MEDAL_PURPLEHEART;
-			give_medal(ob, MEDAL_ID_PURPLEHEART);
-		}
-	}
-
-	if (!(lsc->medals & MEDAL_ACE)) {
-		if (sc->planekills >= 5) {
-			sc->medals |= MEDAL_ACE;
-			give_ribbon(ob, RIBBON_ID_ACE);
-		}
-	}
-
-	if (!(lsc->medals & MEDAL_TOPACE)) {
-		if (sc->planekills >= 25) {
-			sc->medals |= MEDAL_TOPACE;
-			give_ribbon(ob, RIBBON_ID_TOPACE);
-		}
-	}
-
-	if (!(lsc->medals & MEDAL_SERVICE)) {
-		if (sc->killscore >= SERVICE_KILLSCORE) {
-			sc->landings++;
-		}
-
-		if (sc->landings >= 3) {
-			sc->medals |= MEDAL_SERVICE;
-			give_ribbon(ob, RIBBON_ID_SERVICE);
-		}
-	}
-
-	if (sc->killscore >= COMPETENCE_KILLSCORE) {
-		if (!(lsc->medals & MEDAL_COMPETENCE)) {
-			sc->medals |= MEDAL_COMPETENCE;
-			give_medal(ob, MEDAL_ID_COMPETENCE);
-		} else if (!(lsc->medals & MEDAL_COMPETENCE2)) {
-			sc->medals |= MEDAL_COMPETENCE2;
-			give_ribbon(ob, RIBBON_ID_COMPETENCE2);
-		}
-	}
-
-	if (!(lsc->medals & MEDAL_PREVALOUR)) {
-		if (sc->valour >= VALOUR_PRELIMIT) {
-			sc->medals |= MEDAL_PREVALOUR;
-			give_ribbon(ob, RIBBON_ID_PREVALOUR);
-		}
-	}
-
-	if (!(lsc->medals & MEDAL_VALOUR)) {
-		if (sc->valour >= VALOUR_LIMIT) {
-			sc->medals |= MEDAL_VALOUR;
-			give_medal(ob, MEDAL_ID_VALOUR);
-		}
-	}
-
-	sc->killscore = 0;
-	ob->ob_lastscore = ob->ob_score;
-}
-
-static void get_endlevel(OBJECTS *ob)
+static void GiveMedal(OBJECTS *ob, int medal_id)
 {
 	score_t *sc = &ob->ob_score;
-	score_t *lsc = &ob->ob_lastscore;
+	if (!HaveMedal(sc->medals, sc->medals_nr, medal_id)) {
+		assert(sc->medals_nr < 3);
+		sc->medals[sc->medals_nr++] = medal_id;
+	}
+}
 
-	get_awards(ob);
+static void GiveRibbon(OBJECTS *ob, int ribbon_id)
+{
+	score_t *sc = &ob->ob_score;
+	if (!HaveMedal(sc->ribbons, sc->ribbons_nr, ribbon_id)) {
+		assert(sc->ribbons_nr < 6);
+		sc->ribbons[sc->ribbons_nr++] = ribbon_id;
+	}
+}
 
-	if (!(lsc->medals & MEDAL_PERFECT)) {
-		if (ob->ob_crashcnt == 0) {
-			sc->medals |= MEDAL_PERFECT;
-			give_ribbon(ob, RIBBON_ID_PERFECT);
+static void GetAwards(OBJECTS *ob)
+{
+	flight_score_t *fsc = &ob->ob_flightscore;
+	score_t *sc = &ob->ob_score;
+
+	// We only award purple heart if wounded in combat (=hit by bullet),
+	// not from a bird strike or flying into an ox.
+	if (fsc->combatwound) {
+		GiveMedal(ob, MEDAL_PURPLEHEART);
+	}
+
+	// We count up the number of planes shot down, but they only count if
+	// the player returns to base successfully.
+	sc->planekills += fsc->planekills;
+	//printf("planes %d %d\n", sc->planekills, fsc->planekills);
+	if (sc->planekills >= 5) {
+		GiveRibbon(ob, RIBBON_ACE);
+	}
+	if (sc->planekills >= 25) {
+		GiveRibbon(ob, RIBBON_TOPACE);
+	}
+
+	// We count the number of landings during which we did a decent amount
+	// of damage to the enemy.
+	if (fsc->killscore >= SERVICE_KILLSCORE) {
+		sc->landings++;
+	}
+	//printf("kills %d landings %d\n", fsc->killscore, sc->landings);
+	if (sc->landings >= 3) {
+		GiveRibbon(ob, RIBBON_SERVICE);
+	}
+
+	// If a huge number of damage is done, we give a medal (first time) or
+	// ribbon (second time)
+	if (fsc->killscore >= COMPETENCE_KILLSCORE) {
+		if (!HaveMedal(sc->medals, sc->medals_nr, MEDAL_COMPETENCE)) {
+			GiveMedal(ob, MEDAL_COMPETENCE);
+		} else {
+			GiveRibbon(ob, RIBBON_COMPETENCE2);
 		}
 	}
+
+	// We count up valour, but again - only if the player returned to base
+	// successfully
+	sc->valour += fsc->valour;
+	//printf("valour %d %d\n", sc->valour, fsc->valour);
+	if (sc->valour >= VALOUR_PRELIMIT) {
+		GiveRibbon(ob, RIBBON_PREVALOUR);
+	}
+	if (sc->valour >= VALOUR_LIMIT) {
+		GiveMedal(ob, MEDAL_VALOUR);
+	}
+}
+
+static void GetEndLevel(OBJECTS *ob)
+{
+	GetAwards(ob);
+
+	if (ob->ob_crashcnt == 0) {
+		GiveRibbon(ob, RIBBON_PERFECT);
+	}
 }
 
 // plane
@@ -231,22 +227,25 @@ static void get_endlevel(OBJECTS *ob)
 OBJECTS *initpln(OBJECTS * obp)
 {
 	OBJECTS *ob;
-	int x, height, minx, maxx, n;
+	int x, height, minx, maxx;
 
 	if (!obp) {
+		assert(num_planes < MAX_PLYR);
 		ob = allocobj();
+		ob->ob_original_ob = orig_planes[num_planes];
+		planes[num_planes] = ob;
+		++num_planes;
 	} else {
 		ob = obp;
 	}
 
-	n = ob->ob_index;
-	ob->ob_original_ob = &currgame->gm_planes[n];
-
 	if (obp && ob->ob_state != CRASHED && !ob->ob_athome) {
 		/* Just returned home */
-		get_awards(ob);
+		GetAwards(ob);
 	}
 
+	initflightscore(&ob->ob_flightscore);
+
 	ob->ob_type = PLANE;
 
 	ob->ob_x = ob->ob_original_ob->x;
@@ -254,11 +253,11 @@ OBJECTS *initpln(OBJECTS * obp)
 	maxx = ob->ob_x + 20;
 	height = 0;
 	for (x = minx; x <= maxx; ++x) {
-		if (ground[x] > height) {
-			height = ground[x];
-		}
+		height = imax(height, ground[x]);
 	}
 	ob->ob_y = height + 13;
+	ob->ob_orig_y = ob->ob_y;
+
 	ob->ob_lx = ob->ob_ly = ob->ob_speed = ob->ob_flaps = ob->ob_accel
 	    = ob->ob_hitcount = ob->ob_bdelay = ob->ob_mdelay
 	    = ob->ob_bsdelay = 0;
@@ -268,9 +267,10 @@ OBJECTS *initpln(OBJECTS * obp)
 	ob->ob_target = ob->ob_missiletarget = NULL;
 	ob->ob_firing = ob->ob_mfiring = NULL;
 	ob->ob_bombing = ob->ob_bfiring = ob->ob_home = false;
-	ob->ob_newsym = symbol_plane[ob->ob_orient][0];
+	ob->ob_symbol = &symbol_plane[0].sym[ob->ob_orient ? 4 : 0];
 	ob->ob_athome = true;
 	ob->ob_onmap = true;
+	ob->ob_flightscore.combatwound = false;
 
 	if (!obp || ob->ob_state == CRASHED) {
 		/* New plane */
@@ -279,17 +279,14 @@ OBJECTS *initpln(OBJECTS * obp)
 		ob->ob_missiles = MAXMISSILES;
 		ob->ob_bursts = MAXBURSTS;
 		ob->ob_life = MAXFUEL;
-		initplanescore(&ob->ob_score);
-		initplanescore(&ob->ob_lastscore);
 	}
 	if (!obp) {
-		ob->ob_score.score = ob->ob_crashcnt = 0;
+		initscore(&ob->ob_score);
 		ob->ob_endsts = PLAYING;
 		ob->ob_target = NULL;
 		insertx(ob, &topobj);
 	} else {
-		deletex(ob);
-		insertx(ob, ob->ob_xnext);
+		insertx(ob, deletex(ob));
 	}
 
 	ob->ob_state = FLYING;
@@ -298,8 +295,6 @@ OBJECTS *initpln(OBJECTS * obp)
 	return ob;
 }
 
-
-
 // player
 
 void initplyr(OBJECTS * obp)
@@ -308,11 +303,9 @@ void initplyr(OBJECTS * obp)
 
 	ob = initpln(obp);
 	if (!obp) {
-		ob->ob_drawf = dispplyr;
 		ob->ob_movef = moveplyr;
-		ob->ob_clr = ob->ob_index % 2 + 1;
-		ob->ob_owner = ob;
-		oobjects[ob->ob_index] = *ob;
+		ob->ob_faction = ob->ob_original_ob->faction;
+		ob->ob_clr = ob->ob_faction;
 		endcount = 0;
 
 		ob->ob_plrnum = num_players;
@@ -320,7 +313,6 @@ void initplyr(OBJECTS * obp)
 
 		/* todo: save pointers to all player planes
 		 * and turn consoleplayer into a macro */
-
 		if (ob->ob_plrnum == player) {
 			consoleplayer = ob;
 		}
@@ -335,17 +327,11 @@ void initcomp(OBJECTS * obp)
 
 	ob = initpln(obp);
 	if (!obp) {
-		ob->ob_drawf = dispcomp;
 		ob->ob_movef = movecomp;
-		ob->ob_clr = 2;
-		if (playmode != PLAYMODE_ASYNCH) {
-			ob->ob_owner = &nobjects[1];
-		} else if (ob->ob_index == 1) {
-			ob->ob_owner = ob;
-		} else {
-			ob->ob_owner = ob - 2;
-		}
-		oobjects[ob->ob_index] = *ob;
+		// TODO: Allow multiple computer-controlled planes belonging
+		// to different factions.
+		ob->ob_faction = FACTION_PLAYER2;
+		ob->ob_clr = ob->ob_faction;
 	}
 	if (playmode == PLAYMODE_SINGLE || playmode == PLAYMODE_NOVICE) {
 		ob->ob_state = FINISHED;
@@ -354,7 +340,6 @@ void initcomp(OBJECTS * obp)
 	}
 }
 
-
 static int isrange(int x, int y, int ax, int ay)
 {
 	int dx, dy, t;
@@ -376,8 +361,46 @@ static int isrange(int x, int y, int ax, int ay)
 	return (7 * dx + 4 * dy) / 8;
 }
 
-// bullet
+// When targets fire at the player, it is sometimes possible that the bullet's
+// first position is inside the target that fired it. This usually doesn't
+// happen because there is space at the top of the target symbols, but in the
+// DOS Network/Author's Edition versions, hangars can actually blow themselves
+// up by shooting their own flag. When spawning a bullet we therefore perform a
+// check to make sure this isn't going to happen. If we detect a collision, we
+// adjust the bullet spawn position very slightly to avoid it.
+static void AdjustBullet(OBJECTS *bullet, OBJECTS *src)
+{
+	OBJECTS next_bullet, next_src;
+	int x, y;
 
+	copyobj(&next_src, src);
+	movexy(&next_src, &x, &y);
+
+	// We keep adjusting until we don't detect a collision.
+	for (;;) {
+		// Where will the bullet be on the next frame?
+		copyobj(&next_bullet, bullet);
+		movexy(&next_bullet, &x, &y);
+
+		if (!CollisionTest(&next_bullet, &next_src)) {
+			return;
+		}
+
+		if (bullet->ob_dx < 0) {
+			bullet->ob_x -= 2;
+		} else {
+			bullet->ob_x += 2;
+		}
+
+		if (bullet->ob_dy < 0) {
+			bullet->ob_y -= 2;
+		} else {
+			bullet->ob_y += 2;
+		}
+	}
+}
+
+// bullet
 void initshot(OBJECTS *obo, OBJECTS * targ)
 {
 	OBJECTS *ob;
@@ -420,25 +443,31 @@ void initshot(OBJECTS *obo, OBJECTS * targ)
 	}
 
 	ob->ob_type = SHOT;
-	ob->ob_x = obo->ob_x + SYM_WDTH / 2;
-	ob->ob_y = obo->ob_y - SYM_HGHT / 2;
+	ob->ob_x = obo->ob_x + obo->ob_symbol->w / 2;
+	ob->ob_y = obo->ob_y - obo->ob_symbol->h / 2;
 	ob->ob_lx = obo->ob_lx;
 	ob->ob_ly = obo->ob_ly;
 
 	ob->ob_life = BULLIFE;
 	ob->ob_owner = obo;
 	ob->ob_clr = obo->ob_clr;
-	ob->ob_newsym = &symbol_pixel;
-	ob->ob_drawf = NULL;
+	ob->ob_symbol = &symbol_pixel;
+	ob->ob_soundf = NULL;
 	ob->ob_movef = moveshot;
 	ob->ob_speed = 0;
 
-	insertx(ob, obo);
+	// For observation balloons, we adjust the position very slightly,
+	// so that the bullets appear to be being fired from the basket.
+	if (obo->ob_type == BALLOON) {
+		ob->ob_y -= 4;
+	}
 
+	AdjustBullet(ob, obo);
+
+	insertx(ob, obo);
 }
 
 // bomb
-
 void initbomb(OBJECTS *obo)
 {
 	OBJECTS *ob;
@@ -478,16 +507,13 @@ void initbomb(OBJECTS *obo)
 	ob->ob_life = BOMBLIFE;
 	ob->ob_owner = obo;
 	ob->ob_clr = obo->ob_clr;
-	ob->ob_newsym = symbol_bomb[0];
-	ob->ob_drawf = dispbomb;
+	ob->ob_symbol = &symbol_bomb[0].sym[0];
 	ob->ob_movef = movebomb;
 
 	insertx(ob, obo);
-
 }
 
 // missile
-
 void initmiss(OBJECTS *obo)
 {
 	OBJECTS *ob;
@@ -521,20 +547,17 @@ void initmiss(OBJECTS *obo)
 	ob->ob_life = MISSLIFE;
 	ob->ob_owner = obo;
 	ob->ob_clr = obo->ob_clr;
-	ob->ob_newsym = symbol_missile[0];
-	ob->ob_drawf = NULL;
+	ob->ob_symbol = &symbol_missile[0].sym[0];
+	ob->ob_soundf = NULL;
 	ob->ob_movef = movemiss;
 	ob->ob_missiletarget = obo->ob_mfiring;
 	ob->ob_orient = ob->ob_accel = ob->ob_flaps = 0;
 	ob->ob_onmap = true;
 
 	insertx(ob, obo);
-
 }
 
-
 // starburst
-
 void initburst(OBJECTS *obo)
 {
 	OBJECTS *ob;
@@ -575,40 +598,72 @@ void initburst(OBJECTS *obo)
 	ob->ob_life = BURSTLIFE;
 	ob->ob_owner = obo;
 	ob->ob_clr = obo->ob_clr;
-	ob->ob_newsym = symbol_burst[0];
-	ob->ob_drawf = NULL;
+	ob->ob_symbol = &symbol_burst[0].sym[0];
+	ob->ob_soundf = NULL;
 	ob->ob_movef = moveburst;
 
 	insertx(ob, obo);
+}
 
+static void AddPlayerTarget(OBJECTS *ob, const original_ob_t *orig_ob)
+{
+	// TODO: For the time being, player 3 is just interpreted as a
+	// special player number that either maps to player 1 or 2
+	// depending on the game mode; other players are just synonyms
+	// for players 1 and 2.
+	// This may change in the future if we support multiplayer with
+	// more than two players.
+	switch (orig_ob->faction) {
+		case NUM_FACTIONS:
+		case FACTION_NONE:
+			ob->ob_faction = FACTION_NONE;
+			break;
+		case FACTION_PLAYER1:
+		case FACTION_PLAYER5:
+		case FACTION_PLAYER7:
+			ob->ob_faction = FACTION_PLAYER1;
+			break;
+		case FACTION_PLAYER2:
+		case FACTION_PLAYER4:
+		case FACTION_PLAYER6:
+		case FACTION_PLAYER8:
+			ob->ob_faction = FACTION_PLAYER2;
+			break;
+		case FACTION_PLAYER3:
+			if (playmode == PLAYMODE_ASYNCH) {
+				ob->ob_faction = FACTION_PLAYER1;
+			} else {
+				ob->ob_faction = FACTION_PLAYER2;
+			}
+			break;
+	}
+	++numtarg[ob->ob_faction];
 }
 
 // building/target
-
-static OBJECTS *inittarget(original_ob_t *orig_ob)
+static OBJECTS *inittarget(const original_ob_t *orig_ob)
 {
 	OBJECTS *ob;
 	int x;
 	int minh, maxh, aveh, minx, maxx;
 
 	ob = allocobj();
+
+	// Initial symbol:
+	ob->ob_symbol = &symbol_targets[orig_ob->orient].sym[0];
+
 	minx = ob->ob_x = orig_ob->x;
-	maxx = ob->ob_x + 15;
+	maxx = ob->ob_x + ob->ob_symbol->w - 1;
 	minh = 999;
 	maxh = 0;
 	for (x = minx; x <= maxx; ++x) {
-		if (ground[x] > maxh) {
-			maxh = ground[x];
-		}
-		if (ground[x] < minh) {
-			minh = ground[x];
-		}
+		minh = imin(minh, ground[x]);
+		maxh = imax(maxh, ground[x]);
 	}
-	aveh = (minh + maxh) / 2;
 
-	while ((ob->ob_y = aveh + 16) >= MAX_Y) {
-		--aveh;
-	}
+	aveh = (minh + maxh) / 2;
+	aveh = clamp_max(aveh, MAX_Y - ob->ob_symbol->h - 1);
+	ob->ob_y = aveh + ob->ob_symbol->h;
 
 	for (x = minx; x <= maxx; ++x) {
 		ground[x] = aveh;
@@ -619,37 +674,44 @@ static OBJECTS *inittarget(original_ob_t *orig_ob)
 	ob->ob_type = TARGET;
 	ob->ob_state = STANDING;
 	ob->ob_orient = orig_ob->orient;
-
-	switch (orig_ob->owner) {
-		case OWNER_PLAYER1:
-			ob->ob_owner = &nobjects[0];
-			++numtarg[0];
-			break;
-		case OWNER_PLAYER2:
-			ob->ob_owner = &nobjects[1];
-			++numtarg[1];
-			break;
-		case OWNER_PLAYER1_IN_MULT:
-			if (playmode == PLAYMODE_ASYNCH) {
-				ob->ob_owner = &nobjects[0];
-				++numtarg[0];
-			} else {
-				ob->ob_owner = &nobjects[1];
-				++numtarg[1];
-			}
-			break;
-	}
-	ob->ob_clr = ob->ob_owner->ob_clr;
-	ob->ob_newsym = symbol_targets[0];
-	ob->ob_drawf = disptarg;
+	AddPlayerTarget(ob, orig_ob);
+	ob->ob_clr = ob->ob_faction;
 	ob->ob_movef = movetarg;
 	ob->ob_onmap = true;
 
 	return ob;
 }
 
-// explosion
+// Number of explosion fragments is calculated based on the number of pixels
+// in the "destroyed target" sprite:
+static int TargetExplosionSize(int target_type)
+{
+	symset_t *ss = &symbol_target_hit[target_type];
+	sopsym_t *s = &ss->sym[0];
+	int i, num_pixels;
 
+	for (i = 0, num_pixels = 0; i < s->w * s->h; i++) {
+		if (s->data[i] != 0) {
+			++num_pixels;
+		}
+	}
+
+	return clamp_range(10, num_pixels, 64);
+}
+
+// The original Sopwith sprites were all 16x16. Since we now support larger
+// sprites, we apply an adjustment so that entire targets appear to be
+// exploding. This is no-op when the symbol is 16x16 like the original.
+static void ApplySpriteSizeOffset(OBJECTS *ob, OBJECTS *obo, int angle)
+{
+	int rangex = clamp_min(obo->ob_symbol->w, 16) - 16,
+	    rangey = clamp_min(obo->ob_symbol->h, 16) - 16;
+	ob->ob_x += COS(angle) * rangex / (256 * 2);
+	ob->ob_y += SIN(angle) * rangey / (256 * 2);
+	ob->ob_y -= rangey;
+}
+
+// explosion
 void initexpl(OBJECTS *obo, int small)
 {
 	OBJECTS *ob;
@@ -660,14 +722,19 @@ void initexpl(OBJECTS *obo, int small)
 	bool mansym;
 	int orient;
 
-	obox = obo->ob_x + (obo->ob_newsym->w / 2);
-	oboy = obo->ob_y + (obo->ob_newsym->h / 2);
+	obox = obo->ob_x + (obo->ob_symbol->w / 2);
+	oboy = obo->ob_y + (obo->ob_symbol->h / 2);
 	obodx = obo->ob_dx >> 2;
 	obody = obo->ob_dy >> 2;
 	oboclr = obo->ob_clr;
 
 	obotype = obo->ob_type;
-	if (obotype == TARGET && obo->ob_orient == 2) {
+
+	if (obotype == TARGET
+	 && (obo->ob_orient == TARGET_OIL_TANK
+	  || obo->ob_orient == TARGET_TANKER_TRUCK
+	  || obo->ob_orient == TARGET_CUSTOM5
+	  || obo->ob_orient == TARGET_CUSTOM_PASSIVE5)) {
 		ic = 1;
 		// sdh: Oil tank explosions were changed in Sopwith "Network
 		// Edition" to be much less intense; we make this change
@@ -680,8 +747,19 @@ void initexpl(OBJECTS *obo, int small)
 			speed = gminspeed;
 		}
 	} else {
-		ic = small ? 6 : 2;
 		speed = gminspeed >> ((explseed & 7) != 7);
+		if (small) {
+			ic = 6;
+		} else if (obotype == TARGET) {
+			// For targets we calculate the number of fragments
+			// based on the number of pixels in the destroyed-
+			// target sprite. For the original Sopwith destroyed-
+			// building sprite, this evaluates to 2, which is
+			// the original behavior.
+			ic = 110 / TargetExplosionSize(obo->ob_orient);
+		} else {
+			ic = 2;
+		}
 	}
 	mansym = obotype == PLANE
 		 && (obo->ob_state == FLYING || obo->ob_state == WOUNDED);
@@ -700,6 +778,8 @@ void initexpl(OBJECTS *obo, int small)
 
 		ob->ob_x = obox + ob->ob_dx;
 		ob->ob_y = oboy + ob->ob_dy;
+		ApplySpriteSizeOffset(ob, obo, i);
+
 		explseed *= ob->ob_x * ob->ob_y;
 		explseed += 7491;
 		if (!explseed) {
@@ -709,7 +789,8 @@ void initexpl(OBJECTS *obo, int small)
 		ob->ob_life = EXPLLIFE;
 		orient = ob->ob_orient = (explseed & 0x01C0) >> 6;
 		if (mansym && (!orient || orient == 7)) {
-			mansym = orient = ob->ob_orient = 0;
+			orient = ob->ob_orient = 0;
+			mansym = false;
 			ob->ob_dx = obodx;
 			ob->ob_dy = -gminspeed;
 		}
@@ -717,8 +798,7 @@ void initexpl(OBJECTS *obo, int small)
 		ob->ob_lx = ob->ob_ly = ob->ob_hitcount = ob->ob_speed = 0;
 		ob->ob_owner = obo;
 		ob->ob_clr = oboclr;
-		ob->ob_newsym = symbol_debris[0];
-		ob->ob_drawf = dispexpl;
+		ob->ob_symbol = &symbol_debris[0].sym[0];
 		ob->ob_movef = moveexpl;
 
 		if (orient) {
@@ -730,7 +810,6 @@ void initexpl(OBJECTS *obo, int small)
 }
 
 // smoke from falling plane
-
 void initsmok(OBJECTS *obo)
 {
 	OBJECTS *ob;
@@ -749,14 +828,13 @@ void initsmok(OBJECTS *obo)
 	ob->ob_lx = ob->ob_ly = ob->ob_ldx = ob->ob_ldy = 0;
 	ob->ob_life = SMOKELIFE;
 	ob->ob_owner = obo;
-	ob->ob_drawf = NULL;
+	ob->ob_soundf = NULL;
 	ob->ob_movef = movesmok;
 	ob->ob_clr = obo->ob_clr;
 }
 
 // birds
-
-static OBJECTS *initflock(original_ob_t *orig_ob)
+static OBJECTS *initflock(const original_ob_t *orig_ob)
 {
 	OBJECTS *ob;
 	int j;
@@ -778,11 +856,11 @@ static OBJECTS *initflock(original_ob_t *orig_ob)
 	ob->ob_dy = ob->ob_lx = ob->ob_ly = ob->ob_ldx = ob->ob_ldy = 0;
 	ob->ob_orient = 0;
 	ob->ob_life = FLOCKLIFE;
-	ob->ob_owner = ob;
-	ob->ob_newsym = symbol_flock[0];
-	ob->ob_drawf = NULL;
+	ob->ob_faction = FACTION_NONE;
+	ob->ob_symbol = &symbol_flock[0].sym[0];
+	ob->ob_soundf = NULL;
 	ob->ob_movef = moveflck;
-	ob->ob_clr = 9;
+	ob->ob_clr = 1;
 	ob->ob_onmap = true;
 
 	for (j = 0; j < NUM_STRAY_BIRDS; ++j) {
@@ -793,14 +871,13 @@ static OBJECTS *initflock(original_ob_t *orig_ob)
 }
 
 // single bird
-
 void initbird(OBJECTS *obo, int i)
 {
 	OBJECTS *ob;
-	static int ibx[] = { 8, 3, 0, 6, 7, 14, 10, 12 };
-	static int iby[] = { 16, 1, 8, 3, 12, 10, 7, 14 };
-	static int ibdx[] = { -2, 2, -3, 3, -1, 1, 0, 0 };
-	static int ibdy[] = { -1, -2, -1, -2, -1, -2, -1, -2 };
+	static const int ibx[] = { 8, 3, 0, 6, 7, 14, 10, 12 };
+	static const int iby[] = { 16, 1, 8, 3, 12, 10, 7, 14 };
+	static const int ibdx[] = { -2, 2, -3, 3, -1, 1, 0, 0 };
+	static const int ibdy[] = { -1, -2, -1, -2, -1, -2, -1, -2 };
 
 	ob = allocobj();
 	if (!ob) {
@@ -816,17 +893,44 @@ void initbird(OBJECTS *obo, int i)
 	ob->ob_orient = ob->ob_lx = ob->ob_ly = ob->ob_ldx = ob->ob_ldy =
 	    0;
 	ob->ob_life = BIRDLIFE;
-	ob->ob_owner = obo;
-	ob->ob_newsym = symbol_bird[0];
-	ob->ob_drawf = NULL;
+	ob->ob_faction = obo->ob_faction;
+	ob->ob_symbol = &symbol_bird[0].sym[0];
+	ob->ob_soundf = NULL;
 	ob->ob_movef = movebird;
 	ob->ob_clr = obo->ob_clr;
 	insertx(ob, &topobj);
 }
 
-// oxen
+static OBJECTS *initballoon(const original_ob_t *orig_ob)
+{
+	OBJECTS *ob;
 
-static OBJECTS *initox(original_ob_t *orig_ob)
+	ob = allocobj();
+	if (!ob) {
+		return NULL;
+	}
+
+	ob->ob_type = BALLOON;
+	ob->ob_state = FLYING;
+	ob->ob_life = 1;
+	ob->ob_x = orig_ob->x;
+	ob->ob_y = MAX_Y - 16 + SIN(orig_ob->x) / 32;
+	ob->ob_dx = 0;
+	ob->ob_dy = 0;
+	ob->ob_orient = 0;
+	ob->ob_symbol = &symbol_balloon[0].sym[0];
+	ob->ob_soundf = NULL;
+	ob->ob_movef = moveballoon;
+	ob->ob_faction = orig_ob->faction;
+	ob->ob_clr = ob->ob_faction;
+	ob->ob_onmap = true;
+	AddPlayerTarget(ob, orig_ob);
+
+	return ob;
+}
+
+// oxen
+static OBJECTS *initox(const original_ob_t *orig_ob)
 {
 	OBJECTS *ob;
 
@@ -845,9 +949,9 @@ static OBJECTS *initox(original_ob_t *orig_ob)
 	ob->ob_y = ground[ob->ob_x] + 16;
 	ob->ob_orient = ob->ob_lx = ob->ob_ly = ob->ob_ldx =
 	    ob->ob_ldy = ob->ob_dx = ob->ob_dy = 0;
-	ob->ob_owner = ob;
-	ob->ob_newsym = symbol_ox[0];
-	ob->ob_drawf = NULL;
+	ob->ob_faction = FACTION_NONE;
+	ob->ob_symbol = &symbol_ox[0].sym[orig_ob->transform];
+	ob->ob_soundf = NULL;
 	ob->ob_movef = moveox;
 	ob->ob_clr = 1;
 	return ob;
@@ -856,14 +960,13 @@ static OBJECTS *initox(original_ob_t *orig_ob)
 static void inittargets(void)
 {
 	OBJECTS *ob;
-	original_ob_t *orig_ob;
+	const original_ob_t *orig_ob;
 	int i;
 
-	numtarg[0] = 0;
-	numtarg[1] = 0;
+	memset(numtarg, 0, sizeof(numtarg));
 
-	for (i = 0; i < currgame->gm_num_targets; i++) {
-		orig_ob = &currgame->gm_targets[i];
+	for (i = 0; i < currgame->gm_num_objects; i++) {
+		orig_ob = &currgame->gm_objects[i];
 		switch (orig_ob->type) {
 			case TARGET:
 				ob = inittarget(orig_ob);
@@ -874,6 +977,9 @@ static void inittargets(void)
 			case FLOCK:
 				ob = initflock(orig_ob);
 				break;
+			case BALLOON:
+				ob = initballoon(orig_ob);
+				break;
 			default:
 				continue;
 		}
@@ -901,9 +1007,14 @@ void swinitlevel(void)
 {
 	int i;
 
+	if (have_custom_level) {
+		currgame = &custom_level;
+	} else {
+		currgame = &original_level;
+	}
+
 	// clear out any waiting keys. this stops, eg. 's' on the
 	// menu from toggling sound once the game starts
-
 	Vid_GetGameKeys();
 
 	if (playmode == PLAYMODE_ASYNCH) {
@@ -917,6 +1028,23 @@ void swinitlevel(void)
 	initobjs();
 
 	num_players = 0;
+	num_planes = 0;
+	memset(planes, 0, sizeof(planes));
+	memset(orig_planes, 0, sizeof(orig_planes));
+
+	for (i = 0; i < currgame->gm_num_objects; i++) {
+		int pln_index;
+		if (currgame->gm_objects[i].type != PLANE) {
+			continue;
+		}
+		// TODO: Currently we equate faction with plane number,
+		// which is incorrect - it should be possible to have
+		// multiple planes per faction.
+		pln_index = currgame->gm_objects[i].faction - 1;
+		if (in_range(0, pln_index, MAX_PLYR * 2 - 1)) {
+			orig_planes[pln_index] = &currgame->gm_objects[i];
+		}
+	}
 
 	if (keydelay == -1) {
 		keydelay = 1;
@@ -928,17 +1056,19 @@ void swinitlevel(void)
 	} else {
 		maxcrash = MAXCRASH;
 
-		// single player
-		
+		// single player. we spawn as many enemy planes as we have
+		// defined within the map.
+		// TODO: We should have a way to define the number of enemy
+		// planes in levels.
 		initplyr(NULL);
-		initcomp(NULL);
-		initcomp(NULL);
-		initcomp(NULL);
+		for (i = 1; i < MAX_PLYR && orig_planes[i] != NULL; i++) {
+			initcomp(NULL);
+		}
 	}
 
 	inittargets();
 
-	initdisp(NO);
+	initdisp(false);
 	initgdep();
 
 	// sdh 16/11/2001: this needs to be reset with each new game
@@ -959,16 +1089,19 @@ void swrestart(void)
 	int time;
 		
 	if (consoleplayer->ob_endsts == WINNER) {
-		ob = &nobjects[player];
+		ob = planes[player];
 		inc = 0;
 
-		get_endlevel(ob);
+		GetEndLevel(ob);
 
+		// Count down the remaining lives; the player gets awarded
+		// extra points for each life.
 		while (ob->ob_crashcnt < maxcrash) {
 			++ob->ob_crashcnt;
 			inc += 25;
 			ob->ob_score.score += inc;
 
+			swdisp();
 			Vid_Update();
 			
 			time = Timer_GetMS();
@@ -976,19 +1109,20 @@ void swrestart(void)
 		}
 		++gamenum;
 		savescore = ob->ob_score;
-		have_savescore = 1;
+		have_savescore = true;
 	} else {
 		// gamenum = 0;
 		// allow variable start level -- Jesse
 		gamenum = starting_level;
-		have_savescore = 0;
+		have_savescore = false;
 
 		// sh 28/10/2001: go back to the title screen
-
 		playmode = PLAYMODE_UNSET;
 	}
 
-	longjmp(envrestart, 0);
+	// The main loop code in swmain.c will detect this and return
+	// to the title screen.
+	restart_flag = true;
 }
 
 // init game
@@ -1009,33 +1143,50 @@ void swinit(int argc, char *argv[])
 	// longer be necessary.
 	swloadconf();
 
+	// We must generate symbols early, so that custom map files
+	// can replace symbols:
+	GenerateSymbols();
+
 	for (i=1; i<argc; ++i) {
-		if (!strcasecmp(argv[i], "-n")) {
-			n = 1;
+		if (!strcasecmp(argv[i], "-v")
+		 || !strcasecmp(argv[i], "--version")) {
+			puts(PACKAGE_STRING);
+			exit(0);
+		} else if (!strcasecmp(argv[i], "-n")) {
+			n = true;
 		} else if (!strcasecmp(argv[i], "-s")) {
-			s = 1;
+			s = true;
 		} else if (!strcasecmp(argv[i], "-c")) {
-			c = 1;
+			c = true;
 		} else if (!strcasecmp(argv[i], "-f")) {
 			vid_fullscreen = 1;
 		} else if (!strncasecmp(argv[i], "-g", 2)) {
 			sscanf(& (argv[i][2]), "%d", &starting_level);
 			gamenum = starting_level;
 		} else if (!strcasecmp(argv[i], "-q")) {
-			soundflg = 1;
+			soundflg = true;
+		} else if (!strcmp(argv[i], "-m") && (i + 1) < argc) {
+			LoadCustomLevel(argv[i + 1]);
+			++i;
 		} else
 #ifdef TCPIP
 		if (!strcasecmp(argv[i], "-l")) {
-			a = 1;
+			a = true;
 			asynmode = ASYN_LISTEN;
 		} else if (!strcasecmp(argv[i], "-j")) {
 			if (i + 1 >= argc) {
-				error_exit("insufficient arguments to -j");
+				ErrorExit("insufficient arguments to -j");
 			}
-			a = 1;
+			a = true;
 			asynmode = ASYN_CONNECT;
 			++i;
-			snprintf(asynhost, sizeof(asynhost), "%s", argv[i]);
+			asynhost = argv[i];
+		} else if (!strcmp(argv[i], "-p")) {
+			if (i + 1 >= argc) {
+				ErrorExit("insufficient arguments to -p");
+			}
+			++i;
+			asynport = atoi(argv[i]);
 		} else
 #endif
 		{
@@ -1054,27 +1205,18 @@ void swinit(int argc, char *argv[])
 
 	initseed();
 
-	// sdh 27/6/2002: generate symbol objects
-
-	symbol_generate();
-
 	Timer_Init();
 	Vid_Init();
+	Vid_SetVideoPalette(conf_video_palette);
 
-	// dont init speaker if started with -q (quiet)
+	// Don't init speaker if started with -q (quiet)
 	if (soundflg) {
-		Speaker_Init();		// init pc speaker
+		Speaker_Init();
 	}
 
-	// initgrnd() below needs currgame initialized to set the contents of
-	// the ground[] array for the title screen.
-	currgame = &original_level;
-
 	initsndt();
-	initgrnd();
 	
 	// set playmode if we can, from command line options
-
 	playmode =
 		n ? PLAYMODE_NOVICE :
 		s ? PLAYMODE_SINGLE :
